#pragma once

#include <ReferenceCounter.h>
#include <Text.h>

#include "Item.h"

class ItemModifier : public virtual Framework::ReferenceCounter
{
public:
    ItemModifier();
    virtual void applyOn(Item* zItem) = 0;
};

class ConsumeItemModifier : public ItemModifier
{
public:
    ConsumeItemModifier();
    void applyOn(Item* zItem) override;
};

class ConsumeItemModifierFactory
    : public SubTypeFactory<ItemModifier, ConsumeItemModifier>
{
public:
    ConsumeItemModifierFactory();
    ConsumeItemModifier* fromJson(
        Framework::JSON::JSONObject* zJson) const override;
    Framework::JSON::JSONObject* toJsonObject(
        ConsumeItemModifier* zObject) const override;
    JSONObjectValidationBuilder* addToValidator(
        JSONObjectValidationBuilder* builder) const override;
    const char* getTypeToken() const override;
};

class DoNothingModifier : public ItemModifier
{
public:
    DoNothingModifier();
    void applyOn(Item* zItem) override;
};

class DoNothingModifierFactory
    : public SubTypeFactory<ItemModifier, DoNothingModifier>
{
public:
    DoNothingModifierFactory();
    DoNothingModifier* fromJson(
        Framework::JSON::JSONObject* zJson) const override;
    Framework::JSON::JSONObject* toJsonObject(
        DoNothingModifier* zObject) const override;
    JSONObjectValidationBuilder* addToValidator(
        JSONObjectValidationBuilder* builder) const override;
    const char* getTypeToken() const override;
};

class AttributeItemModifier : public ItemModifier
{
private:
    Framework::RCArray<Framework::Text> attributes;
    Framework::RCArray<Framework::JSON::JSONValue> values;

public:
    AttributeItemModifier();
    void applyOn(Item* zItem) override;
    void addAttribute(Framework::Text name, Framework::JSON::JSONValue* value);
    const Framework::RCArray<Framework::Text>& getAttributes() const;
    const Framework::RCArray<Framework::JSON::JSONValue>& getValues() const;
};

class AttributeItemModifierFactory
    : public SubTypeFactory<ItemModifier, AttributeItemModifier>
{
public:
    AttributeItemModifierFactory();
    AttributeItemModifier* fromJson(
        Framework::JSON::JSONObject* zJson) const override;
    Framework::JSON::JSONObject* toJsonObject(
        AttributeItemModifier* zObject) const override;
    JSONObjectValidationBuilder* addToValidator(
        JSONObjectValidationBuilder* builder) const override;
    const char* getTypeToken() const override;
};

class CombinedItemModifier : public ItemModifier
{
    Framework::RCArray<ItemModifier> modifiers;

public:
    CombinedItemModifier();
    void applyOn(Item* zItem) override;
    void addModifier(ItemModifier* zModifier);
    const Framework::RCArray<ItemModifier>& getModifiers() const;
};

class CombinedItemModifierFactory
    : public SubTypeFactory<ItemModifier, CombinedItemModifier>
{
public:
    CombinedItemModifierFactory();
    CombinedItemModifier* fromJson(
        Framework::JSON::JSONObject* zJson) const override;
    Framework::JSON::JSONObject* toJsonObject(
        CombinedItemModifier* zObject) const override;
    JSONObjectValidationBuilder* addToValidator(
        JSONObjectValidationBuilder* builder) const override;
    const char* getTypeToken() const override;
};